#include <SDL/SDL.h>
#include <SDL/SDL_image.h>
#include <SDL/SDL_rotozoom.h>
#include <SDL/SDL_ttf.h>
#include <SDL/SDL_video.h>

#include <cassert>
#include <cmath>
#include <string>
#include <vector>

//#define ENABLE_SCANLINE

const int SCREEN_WIDTH = 640;
const int SCREEN_HEIGHT = 480;
const int SCREEN_BPP = 32;

SDL_Surface *v_logo = NULL;
SDL_Surface *letters = NULL, *lls[5] = {NULL};
SDL_Surface *screen = NULL;
SDL_Surface *message = NULL;
TTF_Font *font = NULL;

int logo_tget_y = 0, logo_y = 0;

static const int lls_pos[][2] = {
    {0, 33}, {33, 55}, {88, 57}, {145, 59}, {205, 65}};

SDL_Surface *load_image(std::string filename) {
    // Temporary storage for the image that's loaded
    SDL_Surface *loadedImage = NULL;
    SDL_Surface *optimizedImage = NULL;

    loadedImage = IMG_Load(filename.c_str());

    if (loadedImage != NULL) {
        optimizedImage = SDL_DisplayFormat(loadedImage);
        SDL_FreeSurface(loadedImage);
    }
    // Return the optimized image
    return optimizedImage;
}

void apply_surface(int x, int y, SDL_Surface *source,
                   SDL_Surface *destination) {
    // Make a temporary rectangle to hold the offsets
    SDL_Rect offset;

    // Give the offsets to the rectangle
    offset.x = x;
    offset.y = y;

    SDL_BlitSurface(source, NULL, destination, &offset);
}

void p(int x, int y, uint8_t r, uint8_t g, uint8_t b) {
    uint8_t *base = &((uint8_t *)screen->pixels)[x * 4 + y * screen->pitch];
    if (x < 0 || x >= SCREEN_WIDTH || y < 0 || y >= SCREEN_HEIGHT) return;
    base[0] = r;
    base[1] = g;
    base[2] = b;
}

float rand_float(float min, float max) {
    float r0 = (float)rand() / (float)RAND_MAX;
    return (r0) * (max - min) + min;
}

void draw_starfield() {
    static constexpr float max_depth = 20;
    static constexpr float num_stars = 200;
    static constexpr float speed = 0.05;
    struct star_t {
        float x, y, z;
    };

    static int c = 0;
    static std::vector<struct star_t> stars;

    if (c == 0) {
        srand(time(NULL));
        for (int i = 0; i < num_stars; i++) {
            float rx = rand_float(-1, 1), ry = rand_float(-1, 1),
                  rz = rand_float(1, max_depth);
            stars.push_back({rx, ry, rz});
        }
    }

    for (auto &i : stars) {
        float k = 0;  //(i.z * i.z / 100.0); aggiungo X
        float px = 0.5 + ((i.x + k) / i.z);
        float py = 0.5 + ((i.y + k) / i.z);
        float c = std::min(255.f, 255.f / float(pow(i.z, 3) * 0.01));
        for (int y = 0; y < 2; y++)
            for (int x = 0; x < 2; x++)
                p(px * screen->w + x, py * screen->h + y, c, c, c);

        i.z -= speed;
        if (i.z < 0.5) i.z += max_depth;
    }

    ++c;
}

struct v_ctx {
    float angle;
    int k, k2;

    v_ctx(int phase) : angle(0), k(-phase), k2(phase) {}
};

void draw_v(v_ctx &ctx, int x, int y) {
    SDL_Rect srect;
    float zoom = 1 + 0.1 * cos(ctx.k2 / 10.f);

    float a2 = sin(ctx.k2 / 10.0) * 5;
    SDL_Surface *surf = rotozoomSurface(v_logo, ctx.angle + a2, zoom, 0);
    srect.x = x - surf->w / 2;
    srect.y = y - surf->h / 2;
    srect.w = surf->w;
    srect.h = surf->h;
    SDL_BlitSurface(surf, NULL, screen, &srect);
    SDL_FreeSurface(surf);

    constexpr int n = 200;  // wait time
    if (ctx.k < n)
        ctx.angle = 0;
    else {
        ctx.angle += ctx.angle / 100.0 + (2.0 + cos((ctx.k - n) / 180.0) * 2.0);
        if (ctx.angle > 360) {
            ctx.angle = ctx.k = 0;
        }
    }

    ++ctx.k;
    ++ctx.k2;
}

struct l_ctx {
    int k;
    float angle = 0;

    l_ctx() : k(0) {}
};

void draw_letters(l_ctx &ctx, int n, int y) {
    SDL_Rect drect;

    int x = SCREEN_WIDTH / 2;

    float zoom = 1.0 + sin(ctx.k / 20.0 + n / 5.0 * M_PI) *
                           (0.03 + sin((ctx.k + n * 10) / 100) * 0.02);
    ctx.angle = 1.0;
    SDL_Surface *surf = rotozoomSurface(lls[n], ctx.angle, zoom, 0);

    drect.x = x - letters->w / 2 + 17 +
              ((n > 0) ? (lls_pos[n - 1][0] + lls_pos[n - 1][1]) : 0);
    drect.y = y - surf->h / 2 + 13;

    drect.w = surf->w;
    drect.h = surf->h;
    SDL_BlitSurface(surf, NULL, screen, &drect);
    SDL_FreeSurface(surf);
    ctx.k++;
}

void draw_v_2() {
    static std::vector<std::pair<int, int>> pos(8);
    static std::vector<v_ctx> ctx;
    static int k = 0;

    if (k == 0) {
        for (int i = 0; i < pos.size(); i++) {
            pos[i] = std::make_pair(((SCREEN_WIDTH + 64) / pos.size()) * i, 0);
            ctx.emplace_back(rand() % 100);
        }
    }

    for (int i = 0; i < pos.size(); i++) {
        draw_v(ctx[i], pos[i].first, SCREEN_HEIGHT / 2 + pos[i].second);
        pos[i].first -= 5;
        pos[i].second = sin(k / 20.f + (i / (float)pos.size() * 6.28)) * 100;
        if (pos[i].first < -64) {
            pos[i].first += SCREEN_WIDTH + 64;
        }
    }
    k++;
}

// clang-format off
const char *msgs[] = {
    " ", 
    "Streaming is starting soon... :)",
    "Technical issues.. just a sec",
    "rev.ng hour will start @ 11 AM GMT+2",
    "Thank you for watching"
};
// clang-format on

void update_text(int n) {
    if (n >= sizeof(msgs) / sizeof(msgs[0]) || n < 0) return;
    if (message != NULL) SDL_FreeSurface(message);
    message =
        ::TTF_RenderText_Solid(font, msgs[n], SDL_Color{255, 255, 255, 255});

    if (n != 0)
        logo_tget_y = 40 + v_logo->h / 2;
    else
        logo_tget_y = SCREEN_HEIGHT / 2;
}

int main() {
    // Initialize all SDL subsystems
    if (SDL_Init(SDL_INIT_EVERYTHING) == -1) {
        printf("Cannot init SDL\n");
        return 1;
    }

    if (!(IMG_Init(IMG_INIT_PNG) & IMG_INIT_PNG)) {
        printf("Cannot init SDL_image (PNG)\n");
        return 1;
    }

    if (TTF_Init() == -1) {
        printf("Cannot init SDL_ttf\n");
        return 1;
    }
    // Set up the screen
    screen = SDL_SetVideoMode(SCREEN_WIDTH, SCREEN_HEIGHT, SCREEN_BPP,
                              SDL_HWSURFACE);

    // If there was an error in setting up the screen
    if (screen == NULL) {
        printf("Cannot init screen\n");
        return 1;
    }

    font = TTF_OpenFont("vcr.ttf", 28);
    if (font == NULL) {
        printf("cannot open vcr.ttf\n");
        return 1;
    }

    // Set the window caption
    SDL_WM_SetCaption("rev.ng demo", NULL);

    // Load the images
    v_logo = load_image("v.png");
    letters = load_image("letters.png");
    for (int i = 0; i < 5; i++) {
        SDL_Rect srect, drect;
        lls[i] = SDL_CreateRGBSurface(
            letters->flags, lls_pos[i][1], letters->h,
            letters->format->BitsPerPixel, letters->format->Rmask,
            letters->format->Gmask, letters->format->Bmask,
            letters->format->Amask);
        drect.x = 0;
        drect.y = 0;
        drect.w = lls_pos[i][1];
        drect.h = letters->h;

        srect.x = lls_pos[i][0];
        srect.y = 0;
        srect.w = lls_pos[i][1];
        srect.h = letters->h;
        SDL_BlitSurface(letters, &srect, lls[i], &drect);
    }

    // update_text after load_image(v.png)
    update_text(0);
    logo_y = logo_tget_y;

    bool stopEvent = false;
    SDL_Event e;
    int k = 0;

    v_ctx vv(0);
    l_ctx ll[5];
    while (!stopEvent) {
        while (SDL_PollEvent(&e)) {
            if (e.type == SDL_QUIT) {
                stopEvent = true;
            }
            if (e.type == SDL_KEYDOWN) {
                switch (e.key.keysym.sym) {
                    case SDLK_ESCAPE:
                        stopEvent = true;
                        break;
                    default:
                        break;
                }
            }
            if (e.type == SDL_KEYUP) {
                if (e.key.keysym.sym >= SDLK_1 && e.key.keysym.sym <= SDLK_9) {
                    update_text(e.key.keysym.sym - SDLK_1);
                }
            }
        }

        memset(screen->pixels, 0, screen->w * screen->h * 4);
        draw_starfield();
        SDL_Rect drect;
        drect.x = SCREEN_WIDTH / 2 - message->w / 2;
        drect.y = SCREEN_HEIGHT - message->h - 40;
        drect.w = message->w;
        drect.h = message->h;
        SDL_BlitSurface(message, NULL, screen, &drect);
        // draw_v_2();
        if (logo_y != logo_tget_y) logo_y = logo_y * 0.9 + logo_tget_y * 0.1;

        for (int n = 0; n < 5; n++) draw_letters(ll[n], n, logo_y);
        draw_v(vv, SCREEN_WIDTH / 2, logo_y);

#ifdef ENABLE_SCANLINE
        for (int i = 0; i < SCREEN_HEIGHT; i += 2) {
            for (int j = 0; j < SCREEN_WIDTH; j++) {
                uint32_t *pos =
                    ((uint32_t *)screen->pixels) + (screen->pitch * i / 4) + j;
                *pos = (*pos >> 1) & 0x7f7f7f7f;
            }
        }
#endif
        if (SDL_Flip(screen) == -1) {
            return 1;
        }
        ++k;
        SDL_Delay(10);
    }

    // Update the screen
    // Free the surfaces
    SDL_FreeSurface(v_logo);
    SDL_FreeSurface(letters);
    for (int i = 0; i < 5; i++) SDL_FreeSurface(lls[i]);

    // Quit SDL
    SDL_Quit();

    // Return
    return 0;
}
